package models.owl;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * Класс расписания.
 * @author M. Navrotskiy
 * @version 1.0
 */
public class Schedule {

    /**
     * Конструктор.
     * @param id Идентификатор.
     * @param owlURI uri экземплра класса в приложении.
     */
    public Schedule (final int id, final String owlURI) {
        this.subjects = new ArrayList<Subject>();
        this.id = id;
        this.owlURI = owlURI;
    }

    /**
     * Метод удаления всех занятий.
     */
    public void clearSubjects() {
        this.subjects.clear();
    }

    /**
     * Метод изменения занятия.
     * @param id Идентификатор.
     * @param subject Занятие.
     */
    public void updateSubject (final int id, Subject subject) {
        this.removeSubject(id);
        this.subjects.add(subject);
    }

    /**
     * Метод удаления занятия.
     * @param id Идентификатор удаляемого занятия.
     */
    public void removeSubject (int id) {
        Iterator<Subject> it = this.subjects.iterator();

        while (it.hasNext()) {
            if (it.next().getId() == id) {
                it.remove();
            }
        }
    }

    /**
     * Метод добавления нового занятия.
     * @param subject Новое занятие.
     */
    public void appendSubject (Subject subject) {
        this.subjects.add(subject);
    }

    /**
     * Метод получения общего количества занятий.
     * @return Общее количество занятий.
     */
    public int getSubjectCount() {
        return this.subjects.size();
    }

    /**
     * Метод получения итератора на список занятий.
     * @return Итератор на список занятий.
     */
    public Iterator<Subject> getSubjectIterator() {
        return this.subjects.iterator();
    }

    /**
     * Метод получения занятие по индексу.
     * @param index Индекс искомого занятия.
     * @return Занятие.
     */
    public Subject getSubjectByIndex (final int index) {
        return this.subjects.get(index);
    }

    /**
     * Метод получения занятия по идентификатору.
     * @param id Идентификатор искомой записи.
     * @return Занятие или null.
     */
    public Subject getSubjectById (final int id) {
        Iterator<Subject> it = this.subjects.iterator();

        while (it.hasNext()) {
            Subject s = it.next();

            if (s.getId() == id) {
                return s;
            }
        }

        return null;
    }

    /**
     * Метод изменения идентификатора.
     * @param id Новый идентификатор.
     */
    public void setId (final int id) {
        this.id = id;
    }

    /**
     * Метод изменения uri экземпляра класса в онтологии.
     * @param owlURI uri экземпляра класса в онтологии.
     */
    public void setOwlURI (final String owlURI) {
        this.owlURI = owlURI;
    }

    /**
     * Метод получения идентификатора.
     * @return Идентификатор.
     */
    public int getId() {
        return this.id;
    }

    /**
     * Метод получения uri экземпляра класса в онтологии.
     * @return uri экземплра класса в онтологии.
     */
    public String getOwlURI() {
        return this.owlURI;
    }

    /* Поля класса. */
    /** Идентификатор расписания. */
    private int id;
    /** Список учебных предметов. */
    private ArrayList<Subject> subjects;
    /** uri экземпляра класса в онтологии. */
    private String owlURI;

}

