package models.owl;

import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;

/**
 * Класс учебного предмета.
 * @author M. Navrotskiy
 * @version 1.0
 */
public class Subject {

    /**
     * Конструктор.
     * @param id Идентификатор.
     * @param name Название.
     * @param flag Лекция это или нет.
     * @param room Аудитория.
     * @param teacher Преподаватель.
     * @param time Пара.
     * @param owlURI uri экземпляра класса в онтологии.
     * @param week Неделя на которой проводится занятие.
     * @param day День на котором проводится занятие.
     */
    public Subject (int id, String name, boolean flag, Room  room,
                    Teacher teacher, int time, String owlURI, int week, int day) {
        this.classes = new ArrayList<Class>();
        this.id = id;
        this.time = time;
        this.teacher = teacher;
        this.room = room;
        this.name = name;
        this.isLection = flag;
        this.owlURI = owlURI;
        this.week = week;
        this.day = day;
    }

    /**
     * Метод удаления всех учебных групп.
     */
    public void clearClasses() {
        this.classes.clear();
    }

    /**
     * Метод изменения учебной группы.
     * @param id Идентификатор изменяемой группы.
     * @param cls Новая группа.
     */
    public void updateClass (int id, Class cls) {
        removeClass(id);
        this.classes.add(cls);
    }

    /**
     * Метод удаления группы из списка по ее идентификатору.
     * @param id Идентификатор удаляемой группы.
     */
    public void removeClass (int id) {

        Iterator<Class> it = this.classes.iterator();

        while (it.hasNext()) {
            Class c = it.next();
            if (c.getId() == id) {
                it.remove();
            }
        }
    }

    /**
     * Метод добавления учебной группы.
     * @param cls Учебная группа.
     */
    public void appendClass (Class cls) {
        this.classes.add(cls);
    }

    /**
     * Метод получения количества учебных групп.
     * @return Количество учебных групп.
     */
    public int getClassCount() {
        return this.classes.size();
    }

    /**
     * Метод получения итератора списка учебных групп.
     * @return Итератор.
     */
    public Iterator<Class> getClassItertor() {
        return this.classes.iterator();
    }

    /**
     * Метод получения учебной группы по индексу в списке.
     * @param index Индекс учебной группы.
     * @return Наденная учебная группа или null.
     */
    public Class getClassByIndex (int index) {
        return this.classes.get(index);
    }

    /**
     * Метод получения группы по ее id;
     * @param id Идентификатор искомой группы.
     * @return Найденная учебная группа или null.
     */
    public Class getClassById (int id) {
        Iterator<Class> it = this.classes.iterator();

        while (it.hasNext()) {
            Class c = it.next();
            if (c.getId() == id) {
                return c;
            }
        }

        return null;
    }

    /**
     * Метод изменения преподавателя.
     * @param teacher Новый преподаватель.
     */
    public void setTeacher (Teacher teacher) {
        this.teacher = teacher;
    }

    /**
     * Метод изменения аудитории.
     * @param room Новая аудитория.
     */
    public void setRoom (Room room) {
        this.room = room;
    }

    /**
     * Метод изменения пары на которой читается предмет.
     * @param time Новая пара.
     */
    public void setTime (int time) {
        this.time = time;
    }

    /**
     * Метод изменения параметра лекционное ли это занятие.
     * @param flag Лекционное ли это занятие.
     */
    public void setLection (boolean flag) {
        this.isLection = flag;
    }

    /**
     * Метод изменения названия.
     * @param name Новое название.
     */
    public void setName (String name) {
        this.name = name;
    }

    /**
     * Метод изменения идентификатора.
     * @param id Новый идентификатор.
     */
    public void setId (int id) {
        this.id = id;
    }

    /**
     * Метод получения преподавателя, которой ведет занятие.
     * @return Преподаватель.
     */
    public Teacher getTeacher() {
        return this.teacher;
    }

    /**
     * Метод получения аудитории, в которой читается предмет.
     * @return Аудитория.
     */
    public Room getRoom() {
        return this.room;
    }

    /**
     * Метод получения пары на которой ведется предмет.
     * @return  Пара на которой ведется занятие.
     */
    public int getTime() {
        return  this.time;
    }

    public String getTimeString() {
        return SubjectTime.toString(this.time);
    }

    public String getDayString() {
        return DaySubject.toString(this.day);
    }

    public String getWeekString() {
        return SubjectWeek.toString(this.week);
    }

    public String getIsLectionString() {

        return this.isLection ? "Да" : "Нет";
    }

    /**
     * Метод получения параметра того, лекция это или нет.
     * @return true, если лекция.
     */
    public boolean isLection() {
        return this.isLection;
    }

    /**
     * Метод получения названия.
     * @return Название.
     */
    public String getName() {
        return this.name;
    }

    /**
     * Метод получения идентификатора.
     * @return Идентификатор.
     */
    public int getId() {
        return this.id;
    }

    /**
     * Метод получения uri экземпляра класса в онтологии
     * @return uri экземпляра класса в онтологии.
     */
    public String getOwlURI() {
        return this.owlURI;
    }

    /**
     * Метод изменения uri экземпляра класса в онтологии.
     * @param owlURI uri экземпляра класса в онтологии.
     */
    public void setOwlURI (String owlURI) {
        this.owlURI = owlURI;
    }

    /**
     * Метод получения недели на которой проводится занятие.
     * @return Значение недели.
     */
    public int getWeek() {
        return this.week;
    }

    /**
     * Метод получения дня недели, на котором проводится занятие.
     * @return Значение дня недели.
     */
    public int getDay() {
        return this.day;
    }

    /**
     * Метод измения дня недели.
     * @param day День неделии.
     */
    public void setDay (int day) {
        this.day = day;
    }

    /**
     * Метод изменения недели.
     * @param week Неделя.
     */
    public void setWeek (int week) {
        this.week = week;
    }

    /* Поля класса. */
    /** Идентификатор предмета. */
    private int id;
    /** Название предмета. */
    private String name;
    /** Параметр лекция ли это. */
    private boolean isLection;
    /** Список групп в которых ведется предмет. */
    private ArrayList<Class> classes;
    /** Пары на которых ведется предмет. */
    private int time;
    /** Аудитория в которой ведется предмет. */
    private Room room;
    /** Преподаватель, читающий предмет. */
    private Teacher teacher;
    /** uri экземпляра класса в онтологии. */
    private String owlURI;
    /** Наделя на котором проводится занятие. */
    private int week;
    /** День недели на котором проводится занятие. */
    private int day;

}

